import json
from typing import Union

from .structs import (
    PublicKeyCredentialCreationOptions,
    PublicKeyCredentialRequestOptions,
)
from .options_to_json_dict import options_to_json_dict


def options_to_json(
    options: Union[
        PublicKeyCredentialCreationOptions,
        PublicKeyCredentialRequestOptions,
    ],
) -> str:
    """
    Convert registration or authentication options into a simple JSON dictionary, and then stringify
    the result to send to the front end as `Content-Type: application/json`. Alternatively use
    `webauthn.helpers.options_to_json_dict` to get a raw `dict` instead to combine the options with
    other data beforehand/encode with a different scheme/etc...
    """
    return json.dumps(options_to_json_dict(options=options))



from unittest import TestCase

from webauthn.helpers.cose import COSEAlgorithmIdentifier
from webauthn.helpers.options_to_json_dict import options_to_json_dict
from webauthn.helpers.structs import (
    AttestationConveyancePreference,
    AuthenticatorAttachment,
    AuthenticatorSelectionCriteria,
    PublicKeyCredentialDescriptor,
    PublicKeyCredentialHint,
    ResidentKeyRequirement,
    UserVerificationRequirement,
)
from webauthn import generate_registration_options, generate_authentication_options


class TestWebAuthnOptionsToJSON(TestCase):
    maxDiff = None

    def test_converts_registration_options_to_JSON(self) -> None:
        options = generate_registration_options(
            rp_id="example.com",
            rp_name="Example Co",
            user_id=bytes([1, 2, 3, 4]),
            user_name="lee",
            user_display_name="Lee",
            attestation=AttestationConveyancePreference.DIRECT,
            authenticator_selection=AuthenticatorSelectionCriteria(
                authenticator_attachment=AuthenticatorAttachment.PLATFORM,
                resident_key=ResidentKeyRequirement.REQUIRED,
            ),
            challenge=b"1234567890",
            exclude_credentials=[
                PublicKeyCredentialDescriptor(id=b"1234567890"),
            ],
            supported_pub_key_algs=[COSEAlgorithmIdentifier.ECDSA_SHA_512],
            timeout=120000,
            hints=[
                PublicKeyCredentialHint.SECURITY_KEY,
                PublicKeyCredentialHint.CLIENT_DEVICE,
                PublicKeyCredentialHint.HYBRID,
            ],
        )

        output = options_to_json_dict(options)

        self.assertEqual(
            output,
            {
                "rp": {"name": "Example Co", "id": "example.com"},
                "user": {
                    "id": "AQIDBA",
                    "name": "lee",
                    "displayName": "Lee",
                },
                "challenge": "MTIzNDU2Nzg5MA",
                "pubKeyCredParams": [{"type": "public-key", "alg": -36}],
                "timeout": 120000,
                "excludeCredentials": [{"type": "public-key", "id": "MTIzNDU2Nzg5MA"}],
                "authenticatorSelection": {
                    "authenticatorAttachment": "platform",
                    "residentKey": "required",
                    "requireResidentKey": True,
                    "userVerification": "preferred",
                },
                "attestation": "direct",
                "hints": ["security-key", "client-device", "hybrid"],
            },
        )

    def test_converts_authentication_options_to_JSON(self) -> None:
        options = generate_authentication_options(
            rp_id="example.com",
            challenge=b"1234567890",
            allow_credentials=[
                PublicKeyCredentialDescriptor(id=b"1234567890"),
            ],
            timeout=120000,
            user_verification=UserVerificationRequirement.DISCOURAGED,
        )

        output = options_to_json_dict(options)

        self.assertEqual(
            output,
            {
                "rpId": "example.com",
                "challenge": "MTIzNDU2Nzg5MA",
                "allowCredentials": [{"type": "public-key", "id": "MTIzNDU2Nzg5MA"}],
                "timeout": 120000,
                "userVerification": "discouraged",
            },
        )

    def test_converts_authentication_options_to_JSON_custom_bytes_encoder(self) -> None:
        options = generate_authentication_options(
            rp_id="example.com",
            challenge=b"1234567890",
            allow_credentials=[
                PublicKeyCredentialDescriptor(id=b"1234567890"),
            ],
            timeout=120000,
            user_verification=UserVerificationRequirement.DISCOURAGED,
        )

        # Return a representation with the bytes left as-is
        output = options_to_json_dict(options, bytes_encoder=lambda x: x)

        self.assertEqual(
            output,
            {
                "rpId": "example.com",
                "challenge": b"1234567890",
                "allowCredentials": [{"type": "public-key", "id": b"1234567890"}],
                "timeout": 120000,
                "userVerification": "discouraged",
            },
        )



from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from .cose import COSEAlgorithmIdentifier
from .exceptions import MLDSANotSupported
from .decode_credential_public_key import DecodedMLDSAPublicKey


class MLDSAPublicKey(DecodedMLDSAPublicKey):
    """
    Something vaguely shaped like other PublicKey classes in cryptography. Going with something
    like this till the cryptography library itself supports PQC directly.
    """

    def __init__(self, decoded_public_key: DecodedMLDSAPublicKey) -> None:
        assert_ml_dsa_dependencies()

        super().__init__(
            kty=decoded_public_key.kty,
            alg=decoded_public_key.alg,
            pub=decoded_public_key.pub,
        )

    def verify(self, signature: bytes, data: bytes) -> None:
        """
        Verify the ML-DSA signature. Raises `cryptography.exceptions.InvalidSignature` to blend in.
        """
        from dilithium_py.ml_dsa import ML_DSA_44, ML_DSA_65, ML_DSA_87

        if self.alg == COSEAlgorithmIdentifier.ML_DSA_44:
            verified = ML_DSA_44.verify(self.pub, data, signature)
        elif self.alg == COSEAlgorithmIdentifier.ML_DSA_65:
            verified = ML_DSA_65.verify(self.pub, data, signature)
        elif self.alg == COSEAlgorithmIdentifier.ML_DSA_87:
            verified = ML_DSA_87.verify(self.pub, data, signature)

        if not verified:
            raise InvalidSignature()

    def public_bytes(self, encoding: Encoding, format: PublicFormat) -> bytes:
        """
        From https://datatracker.ietf.org/doc/draft-ietf-cose-dilithium/09/:

        "The "pub" parameter is the ML-DSA public key, as described in
        Section 5.3 of FIPS-204."

        This method simply returns the bytes, with no support for other encodings or formats.
        Nothing that A) provides attestation, and B) uses PQC for public keys will use this
        method right now.
        """
        return self.pub


def assert_ml_dsa_dependencies() -> None:
    """
    Check that necessary dependencies are present for handling responses containing ML-DSA public
    keys.

    Raises:
        `webauthn.helpers.exceptions.MLDSANotSupported` if those dependencies are missing
    """
    try:
        import dilithium_py
    except Exception:
        raise MLDSANotSupported(
            "Please install https://pypi.org/project/dilithium-py to verify ML-DSA responses with py_webauthn"
        )
